"""
Utility functions.
"""
import asyncio
import collections
import functools
import inspect
import io
import logging
import os
from typing import Set  # noqa

import libnacl
import logbook
import logbook.compat
import logbook.more
# noinspection PyPackageRequirements
import lru
import wrapt

from .key import Key

__all__ = (
    'enable_logging',
    'disable_logging',
    'get_logger',
    'read_key_or_key_file',
    'raise_server_error',
    'randint',
    'ViewIOReader',
    'ViewIOWriter',
    'async_lru_cache',
    'aio_run',
    'aio_run_decorator',
    'aio_run_proxy_decorator',
    'AioRunMixin',
)

_logger_group = logbook.LoggerGroup()
_logger_group.disabled = True
_logger_redirect_handler = logbook.compat.RedirectLoggingHandler()
_logger_convert_level_handler = logbook.compat.LoggingHandler()


def _convert_level(logging_level):
    return _logger_convert_level_handler.convert_level(logging_level)


def enable_logging(level=logbook.WARNING, asyncio_level=None, aiohttp_level=None):
    # Determine levels
    level = logbook.lookup_level(level)
    converted_level = _convert_level(level)
    if asyncio_level is None:
        asyncio_level = converted_level
    else:
        asyncio_level = _convert_level(asyncio_level)
    if aiohttp_level is None:
        aiohttp_level = converted_level
    else:
        aiohttp_level = _convert_level(aiohttp_level)

    # Enable logger group
    _logger_group.disabled = False

    # Enable asyncio debug logging
    os.environ['PYTHONASYNCIODEBUG'] = '1'

    # Redirect asyncio logger
    logger = logging.getLogger('asyncio')
    logger.setLevel(asyncio_level)
    logger.addHandler(_logger_redirect_handler)

    # Redirect aiohttp logger
    logger = logging.getLogger('aiohttp')
    logger.setLevel(aiohttp_level)
    logger.addHandler(_logger_redirect_handler)


def disable_logging():
    # Reset aiohttp logger
    logger = logging.getLogger('aiohttp')
    logger.removeHandler(_logger_redirect_handler)
    logger.setLevel(logging.NOTSET)

    # Reset asyncio logger
    logger = logging.getLogger('asyncio')
    logger.removeHandler(_logger_redirect_handler)
    logger.setLevel(logging.NOTSET)

    # Disable asyncio debug logging
    del os.environ['PYTHONASYNCIODEBUG']

    # Disable logger group
    _logger_group.disabled = True


def get_logger(name=None, level=logbook.NOTSET):
    """
    Return a :class:`logbook.Logger`.

    Arguments:
        - `name`: The name of a specific sub-logger.
    """
    base_name = 'threema.gateway'
    name = base_name if name is None else '.'.join((base_name, name))

    # Create new logger and add to group
    logger = logbook.Logger(name=name, level=level)
    _logger_group.add_logger(logger)
    return logger


# TODO: Raises
def read_key_or_key_file(key, expected_type):
    """
    Decode a hex-encoded key or read it from a file.

    Arguments:
        - `key`: A hex-encoded key or the name of a file which contains
          a key.
        - `expected_type`: One of the types of :class:`Key.Type`.

    Return a:class:`libnacl.public.SecretKey` or
    :class:`libnacl.public.PublicKey` instance.
    """
    # Read key file (if any)
    try:
        with open(key) as file:
            key = file.readline().strip()
    except IOError:
        pass

    # Convert to key instance
    return Key.decode(key, expected_type)


@asyncio.coroutine
def raise_server_error(response, error):
    """
    Raise a :class:`GatewayServerError` exception from a
    HTTP response. Releases the response before raising.


    Arguments:
        - `response`: A :class:`aiohttp.ClientResponse` instance.
        - `error`: The :class:`GatewayServerError`. to instantiate.

    Always raises :class:`GatewayServerError`.
    """
    status = response.status
    yield from response.release()
    raise error(status)


def randint(a, b):
    """
    Return a cryptographically secure random integer N such that
    ``a <= N <= b``.
    """
    n = libnacl.randombytes_uniform(b) + a
    assert a <= n <= b
    return n


# TODO: Document properly
class ViewIOReader(io.RawIOBase):
    def __init__(self, bytes_or_view):
        super().__init__()
        if isinstance(bytes_or_view, bytes):
            bytes_or_view = memoryview(bytes_or_view)
        self._view = bytes_or_view
        self._offset = 0
        self._length = len(self._view)

    # IOBase methods

    def fileno(self):
        raise OSError('No file descriptors used')

    def isatty(self):
        return False

    def readable(self):
        return True

    def readline(self, size=-1):
        raise NotImplementedError

    def readlines(self, hint=-1):
        raise NotImplementedError

    def seek(self, offset, whence=os.SEEK_SET):
        if whence == os.SEEK_SET:
            pass
        elif whence == os.SEEK_CUR:
            offset += self._offset
        elif whence == os.SEEK_END:
            offset = self._length - offset
        else:
            raise ValueError('Invalid whence value')
        if not 0 < offset <= self._length:
            raise ValueError('Offset is greater than view length')
        self._offset = offset
        return offset

    def seekable(self):
        return True

    def tell(self):
        return self._offset

    def writable(self):
        return False

    # RawIOBase methods

    def read(self, size=-1):
        if size == -1:
            return self.readall()
        elif size < 0:
            raise ValueError('Negative size')
        start, end = self._offset, min(self._offset + size, self._length)
        self._offset = end
        return self._view[start:end]

    def readall(self):
        return self.read(self._length - self._offset)

    def readinto(self, b):
        data = self.readall()
        b.extend(data)
        return len(data)

    # Custom methods

    def __len__(self):
        return self._length - self._offset

    def readexactly(self, size):
        data = self.read(size)
        if len(data) < size:
            raise asyncio.IncompleteReadError(data, size)
        else:
            return data


# TODO: Document properly
class ViewIOWriter(io.RawIOBase):
    def __init__(self, bytes_or_views=None):
        super().__init__()
        self._views = []
        self._length = 0
        if bytes_or_views is not None:
            for bytes_or_view in bytes_or_views:
                self.writeexactly(bytes_or_view)

    # IOBase methods

    def fileno(self):
        raise OSError('No file descriptors used')

    def isatty(self):
        return False

    def readable(self):
        return False

    def seekable(self):
        return False

    def writable(self):
        return True

    # RawIOBase methods

    def write(self, bytes_or_view):
        # Convert to memoryview if necessary
        if isinstance(bytes_or_view, bytes):
            bytes_or_view = memoryview(bytes_or_view)

        # Append
        length = len(bytes_or_view)
        self._length += length
        self._views.append(bytes_or_view)
        return length

    def writelines(self, lines):
        raise NotImplementedError

    # Custom methods

    def __radd__(self, other):
        self.extend(other)
        return self

    def __len__(self):
        return self._length

    def getvalue(self):
        return b''.join(self._views)

    # noinspection PyProtectedMember
    def extend(self, other):
        self._views += other._views
        self._length += other._length

    def writeexactly(self, bytes_or_view):
        return self.write(bytes_or_view)


class _HashedSeq(list):
    """
    This class guarantees that hash() will be called no more than once
    per element.  This is important because the lru_cache() will hash
    the key multiple times on a cache miss.
    """
    __slots__ = 'hash_value'

    # noinspection PyMissingConstructor
    def __init__(self, tuple_):
        self[:] = tuple_
        self.hash_value = hash(tuple_)

    def __hash__(self):
        return self.hash_value


# noinspection PyPep8Naming
_CacheInfo = collections.namedtuple(
    'CacheInfo', ('hits', 'misses', 'maxsize', 'currsize'))


def _make_key(
    args, kwargs, typed,
    fast_types={int, str, frozenset, type(None)},
    kwargs_mark=(object(),),
):
    """
    Make a cache key from optionally typed positional and keyword arguments

    The key is constructed in a way that is flat as possible rather than
    as a nested structure that would take more memory.

    If there is only a single argument and its data type is known to cache
    its hash value, then that argument is returned without a wrapper.  This
    saves space and improves lookup speed.
    """
    key = args
    if kwargs:
        sorted_items = sorted(kwargs.items())
        key += kwargs_mark
        for item in sorted_items:
            key += item
    else:
        sorted_items = []
    if typed:
        key += tuple(type(v) for v in args)
        if kwargs:
            key += tuple(type(v) for k, v in sorted_items)
    elif len(key) == 1 and type(key[0]) in fast_types:
        return key[0]
    return _HashedSeq(key)


class _LRUCacheDict(lru.LRUCacheDict):
    def __init__(self, *args, **kwargs):
        self.hits = self.misses = 0
        super().__init__(*args, **kwargs)

    def __len__(self):
        return self.size()

    def info(self):
        """Report cache statistics"""
        return _CacheInfo(self.hits, self.misses, self.max_size, len(self))

    def __getitem__(self, key):
        try:
            item = super().__getitem__(key)
        except KeyError:
            self.misses += 1
            raise
        else:
            self.hits += 1
            return item

    def clear(self):
        super().clear()
        self.hits = self.misses = 0


def async_lru_cache(maxsize=1024, expiration=15 * 60, typed=False):
    """
    Least-recently-used cache decorator for asyncio coroutines.

    If *maxsize* is set to None, the LRU features are disabled and the
    cache can grow without bound.

    If *expiration* is set, cached values will be cleared after
    *expiration* seconds.

    If *typed* is True, arguments of different types will be cached
    separately. For example, f(3.0) and f(3) will be treated as distinct
    calls with distinct results.

    Arguments to the cached function must be hashable.

    View the cache statistics named tuple (hits, misses, maxsize,
    currsize) with f.cache_info().  Clear the cache and statistics
    with f.cache_clear(). Access the underlying function with
    f.__wrapped__.

    See:  http://en.wikipedia.org/wiki/Cache_algorithms#Least_Recently_Used
    """

    def decorating_function(func):
        cache = _LRUCacheDict(max_size=maxsize, expiration=expiration)

        @asyncio.coroutine
        def wrapper(*args, **kwargs):
            # Make cached key
            key = _make_key(args, kwargs, typed)

            # Get from cache
            try:
                return cache[key]
            except KeyError:
                pass

            # Miss, retrieve from coroutine
            value = yield from func(*args, **kwargs)
            cache[key] = value
            return value

        wrapper.cache = cache
        wrapper.cache_info = cache.info
        wrapper.cache_clear = cache.clear

        return functools.update_wrapper(wrapper, func)

    return decorating_function


def aio_run(coroutine, loop=None, close_after_complete=False):
    """
    Decorator to run an asyncio coroutine as a normal blocking
    function.

    Arguments:
        - `coroutine`: The asyncio coroutine or task to be executed.
        - `loop`: An optional :class:`asyncio.AbstractEventLoop`
          subclass instance.
        - `close_after_complete`: Close `loop` after the coroutine
          returned. Defaults to ``False``.

    Returns the result of the asyncio coroutine.

    Example:

    .. code-block::
        @asyncio.coroutine
        def coroutine(timeout):
            yield from asyncio.sleep(timeout)
            return True

        # Call coroutine in a blocking manner
        result = aio_run(coroutine(1.0))
        print(result)
    """

    # Create a new event loop (if required)
    if loop is None:
        loop_ = asyncio.get_event_loop()

        # Closed? Set a new one
        if loop_.is_closed():
            loop_ = asyncio.new_event_loop()
            asyncio.set_event_loop(loop_)
    else:
        loop_ = loop

    # Run the coroutine and get the result
    result = loop_.run_until_complete(coroutine)

    # Close loop (if requested)
    if close_after_complete:
        loop_.close()

    # Return the result
    return result


def aio_run_decorator(loop=None, close_after_complete=False):
    """
    Decorator to run an asyncio coroutine as a normal blocking
    function.

    Arguments:
        - `loop`: An optional :class:`asyncio.AbstractEventLoop`
          subclass instance.
        - `close_after_complete`: Close `loop` after the coroutine
          returned. Defaults to ``False``.

    Returns a decorator to wrap around an asyncio coroutine.

    Example:

    .. code-block::
        @asyncio.coroutine
        def coroutine(timeout):
            yield from asyncio.sleep(timeout)
            return True

        @aio_run_decorator()
        def helper(*args, **kwargs):
            return coroutine(*args, **kwargs)

        # Call coroutine in a blocking manner
        result = helper(timeout=1.0)
        print(result)
    """
    def _decorator(func):
        # Make it a coroutine if it isn't one already
        if not asyncio.iscoroutinefunction(func):
            func = asyncio.coroutine(func)

        def _wrapper(*args, **kwargs):
            return aio_run(
                func(*args, **kwargs),
                loop=loop,
                close_after_complete=close_after_complete,
            )
        return functools.update_wrapper(_wrapper, func)
    return _decorator


def aio_run_proxy_decorator(cls):
    """
    Proxy a publicly accessible class and run all methods marked as
    async inside it (using the class attribute `async_functions`) with
    an event loop to make it appear as a traditional blocking method.

    Arguments:
        - `cls`: A class to be wrapped. The class must inherit
          :class:`AioRunMixin`. The class and all base classes must
          supply a class attribute `async_functions` which is an
          iterable of method names that should appear as traditional
          blocking functions from the outside.

    Returns a class factory.

    .. note:: The `unwrap` property of the resulting instance can be
              used to get the original instance.
    """
    # Ensure each base class has added a class-level iterable of async functions
    async_functions = set()
    for base_class in inspect.getmro(cls)[:-1]:
        try:
            async_functions.update(base_class.__dict__.get('async_functions', None))
        except TypeError:
            message = "Class {} is missing 'async_functions' iterable"
            raise ValueError(message.format(base_class.__name__))

    # Sanity-check
    if not issubclass(cls, AioRunMixin):
        raise TypeError("Class {} did not inherit 'AioRunMixin'".format(
            cls.__name__))

    class _AioRunProxyDecoratorFactory(wrapt.ObjectProxy):
        def __call__(self, *args, **kwargs):
            # Create instance
            instance = cls(*args, **kwargs)

            # Sanity-check
            if not isinstance(instance, AioRunMixin):
                raise TypeError("Class {} did not inherit 'AioRunMixin'".format(
                    cls.__name__))

            # Wrap with proxy (if required)
            if instance.blocking:
                class _AioRunProxy(wrapt.ObjectProxy):
                    @property
                    def unwrap(self):
                        """
                        Get the wrapped instance.
                        """
                        return self.__wrapped__

                # Wrap all async functions with `aio_run`
                for name in async_functions:
                    def _method(instance_, name_, *args_, **kwargs_):
                        method = aio_run_decorator()(getattr(instance_, name_))
                        return method(*args_, **kwargs_)

                    _method = functools.partial(_method, instance, name)
                    setattr(_AioRunProxy, name, _method)

                return _AioRunProxy(instance)
            else:
                return instance

    return _AioRunProxyDecoratorFactory(cls)


class AioRunMixin:
    """
    Must be inherited when using :func:`aio_run_proxy_decorator`.

    Arguments:
        - `blocking`: Switch to turn the blocking API on or off.
    """
    async_functions = set()  # type: Set[str]

    def __init__(self, blocking=False):
        self.blocking = blocking

    @property
    def unwrap(self):
        """
        Get the wrapped instance.
        """
        return self



import asyncio
import asyncio.subprocess
import binascii
import copy
import hashlib
import hmac
import os
import socket
import subprocess
import sys
import time
from contextlib import closing

import aiohttp
import pytest
from aiohttp import web
from aiohttp.web_urldispatcher import UrlDispatcher

import threema.gateway
from threema.gateway import e2e
from threema.gateway.key import Key

_res_path = os.path.normpath(os.path.join(
    os.path.abspath(__file__), os.pardir, 'res'))


class RawMessage(e2e.Message):
    def __init__(self, connection, nonce=None, message=None, **kwargs):
        super().__init__(connection, e2e.Message.Type.text_message, **kwargs)
        self.nonce = nonce
        self.message = message

    @asyncio.coroutine
    def pack(self, writer):
        raise NotImplementedError

    @classmethod
    @asyncio.coroutine
    def unpack(cls, connection, parameters, key_pair, reader):
        raise NotImplementedError

    @asyncio.coroutine
    def send(self, get_data_only=False):
        """
        Send the raw message

        Return the ID of the message.
        """
        # Send message
        if get_data_only:
            return self.nonce, self.message
        else:
            return (yield from self._connection.send_e2e(**{
                'to': self.to_id,
                'nonce': binascii.hexlify(self.nonce).decode(),
                'box': binascii.hexlify(self.message).decode()
            }))


class Server:
    def __init__(self):
        self.threema_jpg = os.path.join(_res_path, 'threema.jpg')
        key = b'4a6a1b34dcef15d43cb74de2fd36091be99fbbaf126d099d47d83d919712c72b'
        self.echoecho_key = key
        self.echoecho_encoded_key = 'public:' + key.decode('ascii')
        decoded_private_key = Key.decode(pytest.msgapi.private, Key.Type.private)
        self.mocking_key = Key.derive_public(decoded_private_key).hex_pk()
        self.blobs = {}
        self.latest_blob_ids = []

        router = UrlDispatcher()
        router.add_route('GET', '/pubkeys/{key}', self.pubkeys)
        router.add_route('GET', '/lookup/phone/{phone}', self.lookup_phone)
        router.add_route('GET', '/lookup/phone_hash/{phone_hash}', self.lookup_phone_hash)
        router.add_route('GET', '/lookup/email/{email}', self.lookup_email)
        router.add_route('GET', '/lookup/email_hash/{email_hash}', self.lookup_email_hash)
        router.add_route('GET', '/capabilities/{id}', self.capabilities)
        router.add_route('GET', '/credits', self.credits)
        router.add_route('POST', '/send_simple', self.send_simple)
        router.add_route('POST', '/send_e2e', self.send_e2e)
        router.add_route('POST', '/upload_blob', self.upload_blob)
        router.add_route('GET', '/blobs/{blob_id}', self.download_blob)
        self.router = router

    @asyncio.coroutine
    def pubkeys(self, request):
        key = request.match_info['key']
        from_, secret = request.GET['from'], request.GET['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif len(key) != 8:
            return web.Response(status=404)
        elif key == 'ECHOECHO':
            return web.Response(body=self.echoecho_key)
        elif key == '*MOCKING':
            return web.Response(body=self.mocking_key)
        return web.Response(status=404)

    @asyncio.coroutine
    def lookup_phone(self, request):
        phone = request.match_info['phone']
        from_, secret = request.GET['from'], request.GET['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif not phone.isdigit():
            return web.Response(status=404)
        elif phone == '44123456789':
            return web.Response(body=b'ECHOECHO')
        return web.Response(status=404)

    @asyncio.coroutine
    def lookup_phone_hash(self, request):
        phone_hash = request.match_info['phone_hash']
        from_, secret = request.GET['from'], request.GET['secret']
        hash_ = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214c'
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif len(phone_hash) % 2 != 0:
            # Note: This status code might not be intended and may change in the future
            return web.Response(status=500)
        elif len(phone_hash) != 64:
            return web.Response(status=400)
        elif phone_hash == hash_:
            return web.Response(body=b'ECHOECHO')
        return web.Response(status=404)

    @asyncio.coroutine
    def lookup_email(self, request):
        email = request.match_info['email']
        from_, secret = request.GET['from'], request.GET['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif email == 'echoecho@example.com':
            return web.Response(body=b'ECHOECHO')
        return web.Response(status=404)

    @asyncio.coroutine
    def lookup_email_hash(self, request):
        email_hash = request.match_info['email_hash']
        from_, secret = request.GET['from'], request.GET['secret']
        hash_ = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669e2'
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif len(email_hash) % 2 != 0:
            # Note: This status code might not be intended and may change in the future
            return web.Response(status=500)
        elif len(email_hash) != 64:
            return web.Response(status=400)
        elif email_hash == hash_:
            return web.Response(body=b'ECHOECHO')
        return web.Response(status=404)

    @asyncio.coroutine
    def capabilities(self, request):
        id_ = request.match_info['id']
        from_, secret = request.GET['from'], request.GET['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif id_ == 'ECHOECHO':
            return web.Response(body=b'text,image,video,file')
        elif id_ == '*MOCKING':
            return web.Response(body=b'text,image,file')
        return web.Response(status=404)

    @asyncio.coroutine
    def credits(self, request):
        from_, secret = request.GET['from'], request.GET['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        return web.Response(body=b'100')

    @asyncio.coroutine
    def send_simple(self, request):
        post = (yield from request.post())

        # Check API identity
        if (post['from'], post['secret']) not in pytest.msgapi.api_identities:
            return web.Response(status=401)

        # Get ID from to, email or phone
        if 'to' in post:
            id_ = post['to']
        elif post.get('email', None) == 'echoecho@example.com':
            id_ = 'ECHOECHO'
        elif post.get('phone', None) == '44123456789':
            id_ = 'ECHOECHO'
        else:
            return web.Response(status=404)

        # Process
        text = post['text']
        if post['from'] == pytest.msgapi.nocredit_id:
            return web.Response(status=402)
        elif id_ != 'ECHOECHO':
            return web.Response(status=400)
        elif len(text) > 3500:
            return web.Response(status=413)
        return web.Response(body=b'0' * 16)

    @asyncio.coroutine
    def send_e2e(self, request):
        post = (yield from request.post())

        # Check API identity
        if (post['from'], post['secret']) not in pytest.msgapi.api_identities:
            return web.Response(status=401)

        # Get ID, nonce and box
        id_ = post['to']
        nonce, box = binascii.unhexlify(post['nonce']), binascii.unhexlify(post['box'])

        # Process
        if post['from'] == pytest.msgapi.nocredit_id:
            return web.Response(status=402)
        elif id_ != 'ECHOECHO':
            return web.Response(status=400)
        elif len(nonce) != 24:
            # Note: This status code might not be intended and may change in the future
            return web.Response(status=400)
        elif len(box) > 4000:
            return web.Response(status=413)
        return web.Response(body=b'1' * 16)

    @asyncio.coroutine
    def upload_blob(self, request):
        try:
            data = (yield from request.post())

            # Check API identity
            api_identity = (request.GET['from'], request.GET['secret'])
            if api_identity not in pytest.msgapi.api_identities:
                return web.Response(status=401)
        except KeyError:
            return web.Response(status=401)

        try:
            # Get blob
            blob = data['blob'].file.read()
        except KeyError:
            # Note: This status code might not be intended and may change in the future
            return web.Response(status=500)

        # Generate ID
        blob_id = hashlib.md5(blob).hexdigest()

        # Process
        if request.GET['from'] == pytest.msgapi.nocredit_id:
            return web.Response(status=402)
        elif len(blob) == 0:
            return web.Response(status=400)
        elif len(blob) > 20 * (2**20):
            return web.Response(status=413)

        # Store blob and return
        self.blobs[blob_id] = blob
        self.latest_blob_ids.append(blob_id)
        return web.Response(body=blob_id.encode())

    @asyncio.coroutine
    def download_blob(self, request):
        blob_id = request.match_info['blob_id']

        # Check API identity
        from_, secret = request.GET['from'], request.GET['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)

        # Get blob
        try:
            blob = self.blobs[blob_id]
        except KeyError:
            return web.Response(status=404)
        else:
            return web.Response(
                body=blob,
                content_type='application/octet-stream'
            )


def pytest_addoption(parser):
    help_ = 'loop: Use a different event loop, supported: asyncio, uvloop'
    parser.addoption("--loop", action="store", help=help_)


def pytest_report_header(config):
    return 'Using event loop: {}'.format(default_event_loop(config=config))


def pytest_namespace():
    private = 'private:dd9413d597092b004fedc4895db978425efa328ba1f1ec6729e46e09231b8a7e'
    public = Key.encode(Key.derive_public(Key.decode(private, Key.Type.private)))
    values = {'msgapi': {
        'cli_path': os.path.join(
            os.path.dirname(__file__),
            '../threema/gateway/bin/gateway_client.py',
        ),
        'cert_path': os.path.join(_res_path, 'cert.pem'),
        'base_url': 'https://msgapi.threema.ch',
        'ip': '127.0.0.1',
        'id': '*MOCKING',
        'secret': 'mock',
        'private': private,
        'public': public,
        'nocredit_id': 'NOCREDIT',
        'noexist_id': '*NOEXIST',
    }}
    values['msgapi']['api_identities'] = {
        (values['msgapi']['id'], values['msgapi']['secret']),
        (values['msgapi']['nocredit_id'], values['msgapi']['secret'])
    }
    return values


def default_event_loop(request=None, config=None):
    if request is not None:
        config = request.config
    loop = config.getoption("--loop")
    if loop == 'uvloop':
        import uvloop
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    else:
        loop = 'asyncio'
    return loop


def unused_tcp_port():
    """
    Find an unused localhost TCP port from 1024-65535 and return it.
    """
    with closing(socket.socket()) as sock:
        sock.bind((pytest.msgapi.ip, 0))
        return sock.getsockname()[1]


def identity():
    return pytest.msgapi.id, pytest.msgapi.secret


@pytest.fixture(scope='module')
def server():
    return Server()


@pytest.fixture(scope='module')
def raw_message():
    return RawMessage


@pytest.fixture(scope='module')
def event_loop(request):
    """
    Create an instance of the requested event loop.
    """
    default_event_loop(request=request)

    # Close previous event loop
    policy = asyncio.get_event_loop_policy()
    policy.get_event_loop().close()

    # Create new event loop
    _event_loop = policy.new_event_loop()
    policy.set_event_loop(_event_loop)

    def fin():
        _event_loop.close()

    # Add finaliser and return new event loop
    request.addfinalizer(fin)
    return _event_loop


@pytest.fixture(scope='module')
def api_server_port():
    return unused_tcp_port()


@pytest.fixture(scope='module')
def api_server(request, event_loop, api_server_port, server):
    port = api_server_port
    app = web.Application(loop=event_loop, router=server.router)
    handler = app.make_handler()

    # Set up server
    coroutine = event_loop.create_server(handler, host=pytest.msgapi.ip, port=port)
    server_ = event_loop.run_until_complete(coroutine)

    def fin():
        event_loop.run_until_complete(handler.finish_connections(1.0))
        server_.close()
        event_loop.run_until_complete(server_.wait_closed())
        event_loop.run_until_complete(app.cleanup())

    request.addfinalizer(fin)


@pytest.fixture(scope='module')
def mock_url(api_server_port):
    """
    Return the URL where the test server can be reached.
    """
    return 'http://{}:{}'.format(pytest.msgapi.ip, api_server_port)


@pytest.fixture(scope='module')
def connection(request, api_server, mock_url):
    # Note: We're not doing anything with the server but obviously the
    # server needs to be started to be able to connect
    connection_ = threema.gateway.Connection(
        identity=pytest.msgapi.id,
        secret=pytest.msgapi.secret,
        key=pytest.msgapi.private
    )

    # Patch URLs
    connection_.urls = {key: value.replace(pytest.msgapi.base_url, mock_url)
                        for key, value in connection_.urls.items()}

    def fin():
        connection_.close()

    request.addfinalizer(fin)
    return connection_


@pytest.fixture(scope='module')
def connection_blocking(request, api_server, mock_url):
    # Note: We're not doing anything with the server but obviously the
    # server needs to be started to be able to connect
    connection_ = threema.gateway.Connection(
        identity=pytest.msgapi.id,
        secret=pytest.msgapi.secret,
        key=pytest.msgapi.private,
        blocking=True,
    )

    # Patch URLs
    connection_.urls = {key: value.replace(pytest.msgapi.base_url, mock_url)
                        for key, value in connection_.urls.items()}

    def fin():
        connection_.close()

    request.addfinalizer(fin)
    return connection_


@pytest.fixture(scope='module')
def invalid_connection(connection):
    invalid_connection_ = copy.copy(connection)
    invalid_connection_.id = pytest.msgapi.noexist_id
    return invalid_connection_


@pytest.fixture(scope='module')
def nocredit_connection(connection):
    nocredit_connection_ = copy.copy(connection)
    nocredit_connection_.id = pytest.msgapi.nocredit_id
    return nocredit_connection_


@pytest.fixture(scope='module')
def blob():
    return b'\x01\x02\x03'


@pytest.fixture(scope='module')
def blob_id(event_loop, connection, blob):
    coroutine = connection.upload(blob)
    return event_loop.run_until_complete(coroutine)


@pytest.fixture(scope='module')
def cli(api_server, api_server_port, event_loop):
    @asyncio.coroutine
    def call_cli(*args, input=None, timeout=3.0):
        # Prepare environment
        env = os.environ.copy()
        env['THREEMA_TEST_API'] = str(api_server_port)
        test_api_mode = 'WARNING: Currently running in test mode!'

        # Call CLI in subprocess and get output
        parameters = [sys.executable, pytest.msgapi.cli_path] + list(args)
        if isinstance(input, str):
            input = input.encode('utf-8')

        # Create process
        create = asyncio.create_subprocess_exec(
            *parameters, env=env, stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT)
        process = yield from create

        # Wait for process to terminate
        coroutine = process.communicate(input=input)
        output, _ = yield from asyncio.wait_for(coroutine, timeout, loop=event_loop)

        # Process output
        output = output.decode('utf-8')
        if test_api_mode not in output:
            print(output)
            raise ValueError('Not running in test mode')

        # Strip leading empty lines and pydev debugger output
        rubbish = [
            'pydev debugger: process',
            'Traceback (most recent call last):',
            test_api_mode,
        ]
        lines = []
        skip_following_empty_lines = True
        for line in output.splitlines(keepends=True):
            if any((line.startswith(s) for s in rubbish)):
                skip_following_empty_lines = True
            elif not skip_following_empty_lines or len(line.strip()) > 0:
                lines.append(line)
                skip_following_empty_lines = False

        # Strip trailing empty lines
        empty_lines_count = 0
        for line in reversed(lines):
            if len(line.strip()) > 0:
                break
            empty_lines_count += 1
        if empty_lines_count > 0:
            lines = lines[:-empty_lines_count]
        output = ''.join(lines)

        # Check return code
        if process.returncode != 0:
            raise subprocess.CalledProcessError(process.returncode, parameters,
                                                output=output)
        return output
    return call_cli


@pytest.fixture(scope='module')
def private_key_file(tmpdir_factory):
    file = tmpdir_factory.mktemp('keys').join('private_key')
    file.write(pytest.msgapi.private)
    return str(file)


@pytest.fixture(scope='module')
def public_key_file(tmpdir_factory):
    file = tmpdir_factory.mktemp('keys').join('public_key')
    file.write(pytest.msgapi.public)
    return str(file)


class Callback(e2e.AbstractCallback):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = asyncio.Queue(loop=self.loop)

    @asyncio.coroutine
    def receive_message(self, message):
        yield from self.queue.put(message)


@pytest.fixture(scope='module')
def callback(event_loop, connection):
    return Callback(connection, loop=event_loop)


@pytest.fixture(scope='module')
def callback_server_port():
    return unused_tcp_port()


@pytest.fixture(scope='module')
def callback_server(request, event_loop, callback, callback_server_port):
    cert_path = pytest.msgapi.cert_path
    server_ = event_loop.run_until_complete(callback.create_server(
        certfile=cert_path, host=pytest.msgapi.ip, port=callback_server_port))

    def fin():
        server_.close()
        event_loop.run_until_complete(server_.wait_closed())
        event_loop.run_until_complete(callback.close())

    request.addfinalizer(fin)


@pytest.fixture(scope='module')
def callback_client(request, event_loop, callback_server):
    # Note: This is ONLY required because we are using a self-signed certificate
    #       for test purposes.
    connector = aiohttp.TCPConnector(verify_ssl=False)
    session = aiohttp.ClientSession(connector=connector, loop=event_loop)

    def fin():
        session.close()

    request.addfinalizer(fin)
    return session


@pytest.fixture(scope='module')
def callback_send(callback_client, callback_server_port, connection):
    @asyncio.coroutine
    def send(message):
        # Get data from message
        nonce, data = yield from message.send(get_data_only=True)

        # Create callback parameters
        params = {
            'from': connection.id,
            'to': message.to_id,
            'messageId': hashlib.md5(message.to_id.encode('ascii')).hexdigest()[16:],
            'date': str(time.time()),
            'nonce': binascii.hexlify(nonce).decode('ascii'),
            'box': binascii.hexlify(data).decode('ascii'),
        }

        # Calculate MAC
        message = ''.join((params['from'], params['to'], params['messageId'],
                           params['date'], params['nonce'], params['box']))
        message = message.encode('ascii')
        encoded_secret = connection.secret.encode('ascii')
        hmac_ = hmac.new(encoded_secret, msg=message, digestmod=hashlib.sha256)
        params['mac'] = hmac_.hexdigest()

        # Send message
        url = 'https://{}:{}/gateway_callback'.format(
            pytest.msgapi.ip, callback_server_port)
        return (yield from callback_client.post(url, data=params))

    return send


@pytest.fixture(scope='module')
def callback_receive(event_loop, callback, callback_server):
    @asyncio.coroutine
    def receive(timeout=3.0):
        coroutine = asyncio.wait_for(callback.queue.get(), timeout, loop=event_loop)
        return (yield from coroutine)

    return receive



"""
The tests provided in this module have been tested for compliance with
the Threema Gateway server. Obviously, the simulated server does not
completely mimic the behaviour of the Gateway server.
"""
import asyncio

import pytest

from threema.gateway import (
    ReceptionCapability,
    e2e,
    simple,
)
from threema.gateway.exception import (
    BlobServerError,
    CreditsServerError,
    IDServerError,
    KeyServerError,
    MessageServerError,
    ReceptionCapabilitiesServerError,
)


@asyncio.coroutine
def get_latest_blob_ids(server, connection):
    blobs = [(yield from (yield from connection.download(blob_id)).read())
             for blob_id in server.latest_blob_ids]
    return blobs


class TestLookupPublicKey:
    @pytest.mark.asyncio
    def test_invalid_identitiy(self, invalid_connection):
        with pytest.raises(KeyServerError) as exc_info:
            yield from invalid_connection.get_public_key('ECHOECHO')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_invalid_length(self, connection):
        with pytest.raises(KeyServerError) as exc_info:
            yield from connection.get_public_key('TEST')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_unknown_id(self, connection):
        with pytest.raises(KeyServerError) as exc_info:
            yield from connection.get_public_key('00000000')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_valid_id(self, connection, server):
        key = yield from connection.get_public_key('ECHOECHO')
        assert key.hex_pk() == server.echoecho_key

    @pytest.mark.asyncio
    def test_cache_expiration(self, connection, server):
        connection.get_public_key.cache_clear()
        connection.get_public_key.cache.expiration = 0.05
        for _ in range(10):
            yield from self.test_valid_id(connection, server)
        cache_info = connection.get_public_key.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 9
        yield from asyncio.sleep(0.2)
        yield from self.test_valid_id(connection, server)
        cache_info = connection.get_public_key.cache_info()
        # For some reason, the cache is not always being cleared and I don't want
        # Travis to fail all the time
        assert cache_info.misses == 1 or cache_info.misses == 2

    @pytest.mark.asyncio
    def test_cache_hits(self, connection, server):
        connection.get_public_key.cache_clear()
        for _ in range(1000):
            yield from self.test_valid_id(connection, server)
        cache_info = connection.get_public_key.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestLookupIDByPhone:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from invalid_connection.get_id(phone='44123456789')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_unknown_phone(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(phone='44987654321')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_invalid_phone(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(phone='-12537825318')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_valid_phone(self, connection):
        id_ = yield from connection.get_id(phone='44123456789')
        assert id_ == 'ECHOECHO'

    @pytest.mark.asyncio
    def test_cache_hits(self, connection):
        connection.get_id.cache_clear()
        for _ in range(1000):
            yield from self.test_valid_phone(connection)
        cache_info = connection.get_id.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestLookupIDByPhoneHash:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from invalid_connection.get_id(phone_hash='invalid_hash')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_invalid_length(self, connection):
        phone_hash = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af14421'
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(phone_hash=phone_hash)
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    def test_odd_length(self, connection):
        phone_hash = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214'
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(phone_hash=phone_hash)
        # Note: This status code might not be intended and may change in the future
        assert exc_info.value.status == 500

    @pytest.mark.asyncio
    def test_invalid_phone_hash(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(phone_hash='1234')
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    def test_unknown_phone_hash(self, connection):
        phone_hash = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214a'
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(phone_hash=phone_hash)
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_valid_phone_hash(self, connection):
        phone_hash = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214c'
        id_ = yield from connection.get_id(phone_hash=phone_hash)
        assert id_ == 'ECHOECHO'

    @pytest.mark.asyncio
    def test_cache_hits(self, connection):
        connection.get_id.cache_clear()
        for _ in range(1000):
            yield from self.test_valid_phone_hash(connection)
        cache_info = connection.get_id.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestLookupIDByEmail:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from invalid_connection.get_id(email='echoecho@example.com')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_unknown_email(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(email='somemail@example.com')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_invalid_email(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(email='invalid')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_valid_email(self, connection):
        id_ = yield from connection.get_id(email='echoecho@example.com')
        assert id_ == 'ECHOECHO'

    @pytest.mark.asyncio
    def test_cache_hits(self, connection):
        connection.get_id.cache_clear()
        for _ in range(1000):
            yield from self.test_valid_email(connection)
        cache_info = connection.get_id.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestLookupIDByEmailHash:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from invalid_connection.get_id(email_hash='invalid_hash')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_invalid_length(self, connection):
        email_hash = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669'
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(email_hash=email_hash)
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    def test_odd_length(self, connection):
        email_hash = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669e'
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(email_hash=email_hash)
        # Note: This status code might not be intended and may change in the future
        assert exc_info.value.status == 500

    @pytest.mark.asyncio
    def test_invalid_email_hash(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(email_hash='1234')
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    def test_unknown_email_hash(self, connection):
        email_hash = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669e1'
        with pytest.raises(IDServerError) as exc_info:
            yield from connection.get_id(email_hash=email_hash)
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_valid_email_hash(self, connection):
        email_hash = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669e2'
        id_ = yield from connection.get_id(email_hash=email_hash)
        assert id_ == 'ECHOECHO'

    @pytest.mark.asyncio
    def test_cache_hits(self, connection):
        connection.get_id.cache_clear()
        for _ in range(1000):
            yield from self.test_valid_email_hash(connection)
        cache_info = connection.get_id.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestReceptionCapabilities:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection):
        with pytest.raises(ReceptionCapabilitiesServerError) as exc_info:
            yield from invalid_connection.get_reception_capabilities('ECHOECHO')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_invalid_length(self, connection):
        with pytest.raises(ReceptionCapabilitiesServerError) as exc_info:
            yield from connection.get_reception_capabilities('TEST')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_unknown_id(self, connection):
        with pytest.raises(ReceptionCapabilitiesServerError) as exc_info:
            yield from connection.get_reception_capabilities('00000000')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_valid_id(self, connection):
        key = yield from connection.get_reception_capabilities('ECHOECHO')
        assert key == {
            ReceptionCapability.text,
            ReceptionCapability.image,
            ReceptionCapability.video,
            ReceptionCapability.file
        }

    @pytest.mark.asyncio
    def test_cache_hits(self, connection):
        connection.get_reception_capabilities.cache_clear()
        for _ in range(1000):
            yield from self.test_valid_id(connection)
        cache_info = connection.get_reception_capabilities.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestCredits:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection):
        with pytest.raises(CreditsServerError) as exc_info:
            yield from invalid_connection.get_credits()
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_valid(self, connection):
        assert (yield from connection.get_credits()) == 100


class TestUploadBlob:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection):
        with pytest.raises(BlobServerError) as exc_info:
            yield from invalid_connection.upload(b'\x01')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_insufficient_credits(self, nocredit_connection):
        with pytest.raises(BlobServerError) as exc_info:
            yield from nocredit_connection.upload(b'\x01')
        assert exc_info.value.status == 402

    @pytest.mark.asyncio
    def test_just_ok(self, connection, server):
        blob_id = yield from connection.upload(bytes(20 * (2**20)))
        assert len(blob_id) == 32
        # Note: Remove big blob because further tests may hang
        del server.blobs[blob_id]

    @pytest.mark.asyncio
    def test_too_big(self, connection, server):
        with pytest.raises(BlobServerError) as exc_info:
            blob_id = yield from connection.upload(bytes((20 * (2**20)) + 1))
            # Note: Remove big blob because further tests may hang
            del server.blobs[blob_id]
        assert exc_info.value.status == 413

    @pytest.mark.asyncio
    def test_zero(self, connection):
        with pytest.raises(BlobServerError) as exc_info:
            yield from connection.upload(b'')
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    def test_file(self, connection):
        assert len((yield from connection.upload(b'\x01'))) == 32


class TestDownloadBlob:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection, blob_id):
        with pytest.raises(BlobServerError) as exc_info:
            yield from (yield from invalid_connection.download(blob_id)).read()
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_invalid_id(self, connection):
        with pytest.raises(BlobServerError) as exc_info:
            yield from (yield from connection.download('f' * 15)).read()
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_unknown_id(self, connection):
        with pytest.raises(BlobServerError) as exc_info:
            yield from (yield from connection.download('f' * 16)).read()
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_file(self, connection, blob_id, blob):
        response = yield from connection.download(blob_id)
        assert (yield from response.read()) == blob

    @pytest.mark.asyncio
    def test_no_credits(self, nocredit_connection, blob_id, blob):
        response = yield from nocredit_connection.download(blob_id)
        assert (yield from response.read()) == blob


class TestSendSimple:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection):
        with pytest.raises(MessageServerError) as exc_info:
            yield from simple.TextMessage(
                connection=invalid_connection,
                to_id='ECHOECHO',
                text='Hello'
            ).send()
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_insufficient_credits(self, nocredit_connection):
        with pytest.raises(MessageServerError) as exc_info:
            yield from simple.TextMessage(
                connection=nocredit_connection,
                to_id='ECHOECHO',
                text='Hello'
            ).send()
        assert exc_info.value.status == 402

    @pytest.mark.asyncio
    def test_message_too_long(self, connection):
        with pytest.raises(MessageServerError) as exc_info:
            yield from simple.TextMessage(
                connection=connection,
                to_id='ECHOECHO',
                text='0' * 3501
            ).send()
        assert exc_info.value.status == 413

    @pytest.mark.asyncio
    def test_unknown_id(self, connection):
        with pytest.raises(MessageServerError) as exc_info:
            yield from simple.TextMessage(
                connection=connection,
                to_id='00000000',
                text='Hello'
            ).send()
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    def test_unknown_email(self, connection):
        with pytest.raises(MessageServerError) as exc_info:
            yield from simple.TextMessage(
                connection=connection,
                email='somemail@example.com',
                text='Hello'
            ).send()
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_unknown_phone(self, connection):
        with pytest.raises(MessageServerError) as exc_info:
            yield from simple.TextMessage(
                connection=connection,
                phone='44987654321',
                text='Hello'
            ).send()
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    def test_via_id(self, connection):
        id_ = yield from simple.TextMessage(
            connection=connection,
            to_id='ECHOECHO',
            text='0' * 3500
        ).send()
        assert id_ == '0' * 16

    @pytest.mark.asyncio
    def test_via_email(self, connection):
        id_ = yield from simple.TextMessage(
            connection=connection,
            email='echoecho@example.com',
            text='Hello'
        ).send()
        assert id_ == '0' * 16

    @pytest.mark.asyncio
    def test_via_phone(self, connection):
        id_ = yield from simple.TextMessage(
            connection=connection,
            phone='44123456789',
            text='Hello'
        ).send()
        assert id_ == '0' * 16


class TestSendE2E:
    @pytest.mark.asyncio
    def test_invalid_identity(self, invalid_connection, server):
        with pytest.raises(MessageServerError) as exc_info:
            yield from e2e.TextMessage(
                connection=invalid_connection,
                to_id='ECHOECHO',
                key=server.echoecho_encoded_key,
                text='Hello'
            ).send()
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    def test_insufficient_credits(self, nocredit_connection, server):
        with pytest.raises(MessageServerError) as exc_info:
            yield from e2e.TextMessage(
                connection=nocredit_connection,
                to_id='ECHOECHO',
                key=server.echoecho_encoded_key,
                text='Hello'
            ).send()
        assert exc_info.value.status == 402

    @pytest.mark.asyncio
    def test_message_too_long(self, connection, raw_message):
        with pytest.raises(MessageServerError) as exc_info:
            yield from raw_message(
                connection=connection,
                to_id='ECHOECHO',
                nonce=b'0' * 24,
                message=b'1' * 4001
            ).send()
        assert exc_info.value.status == 413

    @pytest.mark.asyncio
    def test_unknown_id(self, connection, server):
        with pytest.raises(MessageServerError) as exc_info:
            yield from e2e.TextMessage(
                connection=connection,
                to_id='00000000',
                key=server.echoecho_encoded_key,
                text='Hello'
            ).send()
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    def test_raw(self, connection, raw_message):
        id_ = yield from raw_message(
            connection=connection,
            to_id='ECHOECHO',
            nonce=b'0' * 24,
            message=b'1' * 4000
        ).send()
        assert id_ == '1' * 16

    @pytest.mark.asyncio
    def test_via_id(self, connection):
        id_ = yield from e2e.TextMessage(
            connection=connection,
            to_id='ECHOECHO',
            text='Hello'
        ).send()
        assert id_ == '1' * 16

    @pytest.mark.asyncio
    def test_via_id_and_key(self, connection, server):
        id_ = yield from e2e.TextMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            text='Hello'
        ).send()
        assert id_ == '1' * 16

    @pytest.mark.asyncio
    def test_image(self, connection, server):
        server.latest_blob_ids = []
        id_ = yield from e2e.ImageMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            image_path=server.threema_jpg
        ).send()
        assert id_ == '1' * 16
        assert len(server.latest_blob_ids) == 1
        assert all((yield from get_latest_blob_ids(server, connection)))

    @pytest.mark.asyncio
    def test_file(self, connection, server):
        server.latest_blob_ids = []
        id_ = yield from e2e.FileMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            file_path=server.threema_jpg
        ).send()
        assert id_ == '1' * 16
        assert len(server.latest_blob_ids) == 1
        assert all((yield from get_latest_blob_ids(server, connection)))

    @pytest.mark.asyncio
    def test_file_with_thumbnail(self, connection, server):
        server.latest_blob_ids = []
        id_ = yield from e2e.FileMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            file_path=server.threema_jpg,
            thumbnail_path=server.threema_jpg
        ).send()
        assert id_ == '1' * 16
        assert len(server.latest_blob_ids) == 2
        assert all((yield from get_latest_blob_ids(server, connection)))



"""
Contains functions to decode, encode and generate keys.
"""
import enum
import hashlib
import hmac

import libnacl.encode
import libnacl.public
import libnacl.secret

from .exception import GatewayKeyError

__all__ = (
    'HMAC',
    'Key',
)


class HMAC:
    """
    A collection of HMAC functions used for the gateway service.
    """
    keys = {
        'email': b'\x30\xa5\x50\x0f\xed\x97\x01\xfa\x6d\xef\xdb\x61\x08\x41\x90\x0f'
                 b'\xeb\xb8\xe4\x30\x88\x1f\x7a\xd8\x16\x82\x62\x64\xec\x09\xba\xd7',
        'phone': b'\x85\xad\xf8\x22\x69\x53\xf3\xd9\x6c\xfd\x5d\x09\xbf\x29\x55\x5e'
                 b'\xb9\x55\xfc\xd8\xaa\x5e\xc4\xf9\xfc\xd8\x69\xe2\x58\x37\x07\x23'
    }

    @staticmethod
    def hash(message, hash_type):
        """
        Generate the hash for a message type.

        Arguments:
            - `message`: A message.
            - `hash_type`: `email` or `phone`.

        Return a :class:`hmac.HMAC` instance.
        """
        return hmac.new(HMAC.keys[hash_type], message.encode('ascii'), hashlib.sha256)


class Key:
    """
    Encode or decode a key.
    """
    separator = ':'

    @enum.unique
    class Type(enum.Enum):
        """
        The type of a key.
        """
        private = 'private'
        public = 'public'

    @staticmethod
    def decode(encoded_key, expected_type):
        """
        Decode a key and check its type if required.

        Arguments:
            - `encoded_key`: The encoded key.
            - `expected_type`: One of the types of :class:`Key.Type`.

        Return the key as an :class:`libnacl.public.SecretKey` or
        :class:`libnacl.public.PublicKey` instance.
        """
        # Split key
        try:
            type_, key = encoded_key.split(Key.separator)
        except ValueError as exc:
            raise GatewayKeyError('Invalid key format') from exc
        type_ = Key.Type(type_)

        # Check type
        if type_ != expected_type:
            raise GatewayKeyError('Invalid key type: {}, expected: {}'.format(
                type_, expected_type
            ))

        # De-hexlify
        key = libnacl.encode.hex_decode(key)

        # Convert to SecretKey or PublicKey
        if type_ == Key.Type.private:
            key = libnacl.public.SecretKey(key)
        elif type_ == Key.Type.public:
            key = libnacl.public.PublicKey(key)

        return key

    @staticmethod
    def encode(libnacl_key):
        """
        Encode a key.

        Arguments:
            - `libnacl_key`: An instance of either a
              :class:`libnacl.public.SecretKey` or a
              :class:`libnacl.public.PublicKey`.

        Return the encoded key.
        """
        # Detect key type and hexlify
        if isinstance(libnacl_key, libnacl.public.SecretKey):
            type_ = Key.Type.private
            key = libnacl_key.hex_sk()
        elif isinstance(libnacl_key, libnacl.public.PublicKey):
            type_ = Key.Type.public
            key = libnacl.encode.hex_encode(libnacl_key.pk)
        else:
            raise GatewayKeyError('Unknown key type: {}'.format(libnacl_key))

        # Encode key
        return Key.separator.join((type_.value, key.decode('utf-8')))

    @staticmethod
    def generate_pair():
        """
        Generate a new key pair.

        Return the key pair as a tuple of a
        :class:`libnacl.public.SecretKey` instance and a
        :class:`libnacl.public.PublicKey` instance.
        """
        private_key = libnacl.public.SecretKey()
        public_key = libnacl.public.PublicKey(private_key.pk)
        return private_key, public_key

    @staticmethod
    def generate_secret_key():
        """
        Generate a new secret key box.

        Return a tuple of the key's :class:`bytes` and hex-encoded
        representation.
        """
        box = libnacl.secret.SecretBox()
        return box.sk, box.hex_sk()

    @staticmethod
    def derive_public(private_key):
        """
        Derive a public key from a class:`libnacl.public.SecretKey`
        instance.

        Arguments:
            - `private_key`: A class:`libnacl.public.SecretKey`
              instance.

        Return the :class:`libnacl.public.PublicKey` instance.
        """
        return libnacl.public.PublicKey(private_key.pk)



from threema.gateway import (
    e2e,
    simple,
)


def test_lookup_id_by_phone(connection_blocking):
    identity = connection_blocking.get_id(phone='44123456789')
    assert identity == 'ECHOECHO'


def test_lookup_id_by_phone_hash(connection_blocking):
    hash_ = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214c'
    identity = connection_blocking.get_id(phone_hash=hash_)
    assert identity == 'ECHOECHO'


def test_lookup_public_key(connection_blocking, server):
    key = connection_blocking.get_public_key('ECHOECHO')
    assert key.hex_pk() == server.echoecho_key


def test_lookup_reception_capabilities(connection_blocking):
    capabilities = connection_blocking.get_reception_capabilities('ECHOECHO')
    assert len(capabilities) == 4


def test_send_e2e_text_message(connection_blocking):
    message = e2e.TextMessage(
        connection=connection_blocking,
        to_id='ECHOECHO',
        text='Hello. This works quite nicely!',
    )
    id_ = message.send()
    assert id_ == '1' * 16


def test_send_simple_text_message(connection_blocking):
    message = simple.TextMessage(
        connection=connection_blocking,
        to_id='ECHOECHO',
        text='Hello. This works quite nicely!',
    )
    id_ = message.send()
    assert id_ == '0' * 16



"""
You can modify and use one of the functions below to test the gateway
service with your end-to-end account.
"""
import logbook
import logbook.more

from threema.gateway import (
    Connection,
    GatewayError,
    util,
)
from threema.gateway.e2e import (
    FileMessage,
    ImageMessage,
    TextMessage,
)


def send(connection):
    """
    Send a message to a specific Threema ID.

    Note that the public key will be automatically fetched from the
    Threema servers. It is strongly recommended that you cache
    public keys to avoid querying the API for each message.
    """
    message = TextMessage(
        connection=connection,
        to_id='ECHOECHO',
        text='私はガラスを食べられます。それは私を傷つけません。'
    )
    return message.send()


def send_cached_key(connection):
    """
    Send a message to a specific Threema ID with an already cached
    public key of that recipient.
    """
    message = TextMessage(
        connection=connection,
        to_id='ECHOECHO',
        key='public:4a6a1b34dcef15d43cb74de2fd36091be99fbbaf126d099d47d83d919712c72b',
        text='私はガラスを食べられます。それは私を傷つけません。'
    )
    return message.send()


def send_cached_key_file(connection):
    """
    Send a message to a specific Threema ID with an already cached
    public key (stored in a file) of that recipient.
    """
    message = TextMessage(
        connection=connection,
        to_id='ECHOECHO',
        key_file='ECHOECHO.txt',
        text='私はガラスを食べられます。それは私を傷つけません。'
    )
    return message.send()


def send_image(connection):
    """
    Send an image to a specific Threema ID.

    Note that the public key will be automatically fetched from the
    Threema servers. It is strongly recommended that you cache
    public keys to avoid querying the API for each message.
    """
    message = ImageMessage(
        connection=connection,
        to_id='ECHOECHO',
        image_path='res/threema.jpg'
    )
    return message.send()


def send_file(connection):
    """
    Send a file to a specific Threema ID.

    Note that the public key will be automatically fetched from the
    Threema servers. It is strongly recommended that you cache
    public keys to avoid querying the API for each message.
    """
    message = FileMessage(
        connection=connection,
        to_id='ECHOECHO',
        file_path='res/some_file.zip'
    )
    return message.send()


def send_file_with_thumbnail(connection):
    """
    Send a file to a specific Threema ID including a thumbnail.

    Note that the public key will be automatically fetched from the
    Threema servers. It is strongly recommended that you cache
    public keys to avoid querying the API for each message.
    """
    message = FileMessage(
        connection=connection,
        to_id='ECHOECHO',
        file_path='res/some_file.zip',
        thumbnail_path='res/some_file_thumb.png'
    )
    return message.send()


def main():
    connection = Connection(
        identity='*YOUR_GATEWAY_THREEMA_ID',
        secret='YOUR_GATEWAY_THREEMA_ID_SECRET',
        key='private:YOUR_PRIVATE_KEY',
        verify_fingerprint=True,
        blocking=True,
    )
    try:
        with connection:
            send(connection)
            send_cached_key(connection)
            send_cached_key_file(connection)
            send_image(connection)
            send_file(connection)
            send_file_with_thumbnail(connection)
    except GatewayError as exc:
        print('Error:', exc)


if __name__ == '__main__':
    util.enable_logging(logbook.WARNING)
    log_handler = logbook.more.ColorizedStderrHandler()
    with log_handler.applicationbound():
        main()



"""
The command line interface for the Threema Gateway Callback Server.
"""
import asyncio
import functools

import click
import logbook
import logbook.more

from threema.gateway import __version__ as _version
from threema.gateway import (
    Connection,
    util,
)
from threema.gateway.e2e import AbstractCallback
from threema.gateway.key import Key

_logging_handler = None
_logging_levels = {
    1: logbook.CRITICAL,
    2: logbook.ERROR,
    3: logbook.WARNING,
    4: logbook.NOTICE,
    5: logbook.INFO,
    6: logbook.DEBUG,
    7: logbook.TRACE,
}


class Callback(AbstractCallback):
    @asyncio.coroutine
    def receive_message(self, message):
        click.echo('Got message ({}): {}'.format(repr(message), message))


def aio_serve(close_func):
    loop = asyncio.get_event_loop()

    def decorator(func):
        func = asyncio.coroutine(func)

        def wrapper(*args, **kwargs):
            # Start
            click.echo('Starting')
            open_result = loop.run_until_complete(func(*args, **kwargs))
            click.echo('Started')
            try:
                loop.run_forever()
            except KeyboardInterrupt:
                pass
            click.echo('Closing')
            close_result = loop.run_until_complete(close_func(open_result))
            loop.close()
            click.echo('Closed')
            return open_result, close_result

        return functools.update_wrapper(wrapper, func)

    return decorator


@click.group()
@click.option('-v', '--verbosity', type=click.IntRange(0, len(_logging_levels)),
              default=0, help="Logging verbosity.")
@click.option('-c', '--colored', is_flag=True, help='Colourise logging output.')
@click.pass_context
def cli(ctx, verbosity, colored):
    """
    Command Line Interface. Use --help for details.
    """
    if verbosity > 0:
        # Enable logging
        util.enable_logging(level=_logging_levels[verbosity])

        # Get handler class
        if colored:
            handler_class = logbook.more.ColorizedStderrHandler
        else:
            handler_class = logbook.StderrHandler

        # Set up logging handler
        handler = handler_class(level=_logging_levels[verbosity])
        handler.push_application()
        global _logging_handler
        _logging_handler = handler

    # Create context object
    ctx.obj = {}


@cli.command(short_help='Show version information.', help="""
Show the current version of the Threema Gateway Callback Server.
""")
def version():
    click.echo('Version: {}'.format(_version))


@asyncio.coroutine
def close_server(server_and_callback):
    server, callback = server_and_callback
    server.close()
    yield from server.wait_closed()
    yield from callback.close()


@cli.command(short_help='Start the callback server.', help="""
Start the Threema Gateway Callback Server.
FROM is the API identity and SECRET is the API secret.
CERT represents the path to a file in PEM format containing the SSL
certificate of the server.""")
@click.argument('identity')
@click.argument('secret')
@click.argument('private_key')
@click.argument('cert', type=click.Path(exists=True))
@click.option('-k', '--keyfile', type=click.Path(exists=True), help="""
Path to a file that contains the private key. Will be read from
CERTFILE if not present.""")
@click.option('-h', '--host', help='Bind to a specific host.')
@click.option('-p', '--port', default=443, help='Listen on a specific port.')
@aio_serve(close_server)
def serve(**arguments):
    # Get arguments
    identity = arguments['identity']
    secret = arguments['secret']
    private_key = util.read_key_or_key_file(arguments['private_key'], Key.Type.private)
    certfile = arguments['cert']
    keyfile = arguments.get('keyfile')
    host = arguments.get('host')
    port = arguments['port']

    # Create connection and callback instances
    connection = Connection(identity=identity, secret=secret, key=private_key)
    callback = Callback(connection)

    # Create server
    coroutine = callback.create_server(certfile, keyfile=keyfile, host=host, port=port)
    server = yield from coroutine
    return server, callback


def main():
    try:
        cli()
    except Exception as exc:
        click.echo('An error occurred:', err=True)
        click.echo(exc, err=True)
        raise
    finally:
        if _logging_handler is not None:
            _logging_handler.pop_application()


if __name__ == '__main__':
    main()



import subprocess

import pytest

from threema.gateway import __version__ as _version
from threema.gateway import (
    ReceptionCapability,
    feature_level,
)
from threema.gateway.key import Key


class TestCLI:
    @pytest.mark.asyncio
    def test_invalid_command(self, cli):
        with pytest.raises(subprocess.CalledProcessError):
            yield from cli('meow')

    @pytest.mark.asyncio
    def test_get_version(self, cli):
        output = yield from cli('version')
        assert 'Version: {}'.format(_version) in output
        assert 'Feature Level: {}'.format(feature_level) in output

    @pytest.mark.asyncio
    def test_invalid_key(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli('encrypt', 'meow', 'meow', input='meow')
        assert 'Invalid key format' in exc_info.value.output
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'encrypt', pytest.msgapi.public, pytest.msgapi.private, input='meow')
        assert 'Invalid key type' in exc_info.value.output

    @pytest.mark.asyncio
    def test_encrypt_decrypt(self, cli):
        input = '私はガラスを食べられます。それは私を傷つけません。'
        output = yield from cli(
            'encrypt', pytest.msgapi.private, pytest.msgapi.public, input=input)
        nonce, data = output.splitlines()
        output = yield from cli(
            'decrypt', pytest.msgapi.private, pytest.msgapi.public, nonce, input=data)
        assert input in output

    @pytest.mark.asyncio
    def test_encrypt_decrypt_by_file(self, cli, private_key_file, public_key_file):
        input = '私はガラスを食べられます。それは私を傷つけません。'
        output = yield from cli(
            'encrypt', private_key_file, public_key_file, input=input)
        nonce, data = output.splitlines()
        output = yield from cli(
            'decrypt', private_key_file, public_key_file, nonce, input=data)
        assert input in output

    @pytest.mark.asyncio
    def test_generate(self, cli, tmpdir):
        private_key_file = tmpdir.join('tmp_private_key')
        public_key_file = tmpdir.join('tmp_public_key')
        yield from cli('generate', str(private_key_file), str(public_key_file))
        private_key = Key.decode(private_key_file.read().strip(), Key.Type.private)
        public_key = Key.decode(public_key_file.read().strip(), Key.Type.public)
        assert private_key
        assert public_key

    @pytest.mark.asyncio
    def test_hash_no_option(self, cli):
        with pytest.raises(subprocess.CalledProcessError):
            yield from cli('hash')

    @pytest.mark.asyncio
    def test_hash_valid_email(self, cli):
        hash_ = '1ea093239cc5f0e1b6ec81b866265b921f26dc4033025410063309f4d1a8ee2c'
        output = yield from cli('hash', '-e', 'test@threema.ch')
        assert hash_ in output
        output = yield from cli('hash', '--email', 'test@threema.ch')
        assert hash_ in output

    @pytest.mark.asyncio
    def test_hash_valid_phone_number(self, cli):
        hash_ = 'ad398f4d7ebe63c6550a486cc6e07f9baa09bd9d8b3d8cb9d9be106d35a7fdbc'
        output = yield from cli('hash', '-p', '41791234567')
        assert hash_ in output
        output = yield from cli('hash', '--phone', '41791234567')
        assert hash_ in output

    @pytest.mark.asyncio
    def test_derive(self, cli):
        output = yield from cli('derive', pytest.msgapi.private)
        assert pytest.msgapi.public in output

    @pytest.mark.asyncio
    def test_send_simple(self, cli):
        id_, secret = pytest.msgapi.id, pytest.msgapi.secret
        output = yield from cli('send_simple', 'ECHOECHO', id_, secret, input='Hello!')
        assert output

    @pytest.mark.asyncio
    def test_send_e2e(self, cli, server):
        output_1 = yield from cli(
            'send_e2e', 'ECHOECHO', pytest.msgapi.id, pytest.msgapi.secret,
            pytest.msgapi.private, input='Hello!')
        assert output_1
        output_2 = yield from cli(
            'send_e2e', 'ECHOECHO', pytest.msgapi.id, pytest.msgapi.secret,
            pytest.msgapi.private, '-k', server.echoecho_encoded_key, input='Hello!')
        assert output_2
        assert output_1 == output_2

    @pytest.mark.asyncio
    def test_send_image(self, cli, server):
        server.latest_blob_ids = []
        output_1 = yield from cli(
            'send_image', 'ECHOECHO', pytest.msgapi.id, pytest.msgapi.secret,
            pytest.msgapi.private, server.threema_jpg)
        assert output_1
        assert len(server.latest_blob_ids) == 1
        output_2 = yield from cli(
            'send_image', 'ECHOECHO', pytest.msgapi.id, pytest.msgapi.secret,
            pytest.msgapi.private, server.threema_jpg, '-k', server.echoecho_encoded_key)
        assert output_2
        assert output_1 == output_2
        assert len(server.latest_blob_ids) == 2

    @pytest.mark.asyncio
    def test_send_file(self, cli, server):
        server.latest_blob_ids = []
        output_1 = yield from cli(
            'send_file', 'ECHOECHO', pytest.msgapi.id, pytest.msgapi.secret,
            pytest.msgapi.private, server.threema_jpg)
        assert output_1
        assert len(server.latest_blob_ids) == 1
        output_2 = yield from cli(
            'send_file', 'ECHOECHO', pytest.msgapi.id, pytest.msgapi.secret,
            pytest.msgapi.private, server.threema_jpg, '-k', server.echoecho_encoded_key)
        assert output_2
        assert output_1 == output_2
        assert len(server.latest_blob_ids) == 2
        output = yield from cli(
            'send_file', 'ECHOECHO', pytest.msgapi.id, pytest.msgapi.secret,
            pytest.msgapi.private, server.threema_jpg, '-t', server.threema_jpg)
        assert output
        assert len(server.latest_blob_ids) == 4
        output = yield from cli(
            'send_file', 'ECHOECHO', pytest.msgapi.id, pytest.msgapi.secret,
            pytest.msgapi.private, server.threema_jpg, '-k', server.echoecho_encoded_key,
            '-t', server.threema_jpg)
        assert output
        assert len(server.latest_blob_ids) == 6

    @pytest.mark.asyncio
    def test_lookup_no_option(self, cli):
        with pytest.raises(subprocess.CalledProcessError):
            yield from cli('lookup', pytest.msgapi.id, pytest.msgapi.secret)

    @pytest.mark.asyncio
    def test_lookup_id_by_email(self, cli):
        output = yield from cli(
            'lookup', pytest.msgapi.id, pytest.msgapi.secret,
            '-e', 'echoecho@example.com')
        assert 'ECHOECHO' in output
        output = yield from cli(
            'lookup', pytest.msgapi.id, pytest.msgapi.secret,
            '--email', 'echoecho@example.com')
        assert 'ECHOECHO' in output

    @pytest.mark.asyncio
    def test_lookup_id_by_phone(self, cli):
        output = yield from cli(
            'lookup', pytest.msgapi.id, pytest.msgapi.secret, '-p', '44123456789')
        assert 'ECHOECHO' in output
        output = yield from cli(
            'lookup', pytest.msgapi.id, pytest.msgapi.secret, '--phone', '44123456789')
        assert 'ECHOECHO' in output

    @pytest.mark.asyncio
    def test_lookup_pk_by_id(self, cli, server):
        output = yield from cli(
            'lookup', pytest.msgapi.id, pytest.msgapi.secret, '-i', 'ECHOECHO')
        assert server.echoecho_encoded_key in output
        output = yield from cli(
            'lookup', pytest.msgapi.id, pytest.msgapi.secret, '--id', 'ECHOECHO')
        assert server.echoecho_encoded_key in output

    @pytest.mark.asyncio
    def test_capabilities(self, cli):
        output = yield from cli(
            'capabilities', pytest.msgapi.id, pytest.msgapi.secret, 'ECHOECHO')
        capabilities = {
            ReceptionCapability.text,
            ReceptionCapability.image,
            ReceptionCapability.video,
            ReceptionCapability.file
        }
        assert all((capability.value in output for capability in capabilities))

    @pytest.mark.asyncio
    def test_credits(self, cli):
        output = yield from cli('credits', pytest.msgapi.id, pytest.msgapi.secret)
        assert '100' in output
        output = yield from cli(
            'credits', pytest.msgapi.nocredit_id, pytest.msgapi.secret)
        assert '0' in output

    @pytest.mark.asyncio
    def test_invalid_id(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'credits', pytest.msgapi.noexist_id, pytest.msgapi.secret)
        assert 'API identity or secret incorrect' in exc_info.value.output

    @pytest.mark.asyncio
    def test_insufficient_credits(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            id_, secret = pytest.msgapi.nocredit_id, pytest.msgapi.secret
            yield from cli('send_simple', 'ECHOECHO', id_, secret, input='!')
        assert 'Insufficient credits' in exc_info.value.output



import asyncio

import logbook
import logbook.more

from threema.gateway import (
    Connection,
    util,
)
from threema.gateway.e2e import AbstractCallback


class Callback(AbstractCallback):
    @asyncio.coroutine
    def receive_message(self, message):
        print('Got message ({}): {}'.format(repr(message), message))


@asyncio.coroutine
def start():
    # Create connection instance
    connection = Connection(
        identity='*YOUR_GATEWAY_THREEMA_ID',
        secret='YOUR_GATEWAY_THREEMA_ID_SECRET',
        key='private:YOUR_PRIVATE_KEY'
    )

    # Create the callback instance
    callback = Callback(connection)

    # Start the callback server and listen on any interface at port 8443
    server = yield from callback.create_server(
        certfile='PATH_TO_SSL_PEM_CERTIFICATE',
        keyfile='PATH_TO_SSL_PRIVATE_KEY',
        port=8443
    )

    # Return both server and callback instance
    return server, callback


@asyncio.coroutine
def stop(server, callback):
    server.close()
    yield from server.wait_closed()
    yield from callback.close()


if __name__ == '__main__':
    util.enable_logging(logbook.WARNING)
    log_handler = logbook.more.ColorizedStderrHandler()
    with log_handler.applicationbound():
        loop = asyncio.get_event_loop()
        # Start server
        server, callback = loop.run_until_complete(start())
        # Ctrl+C: Terminate the server and the callback application
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            pass
        loop.run_until_complete(stop(server, callback))
        loop.close()



"""
Provides classes for the simple mode.
"""
import abc
import asyncio

from .exception import MessageError
from .util import (
    AioRunMixin,
    aio_run_proxy_decorator,
)

__all__ = (
    'Message',
    'TextMessage',
)


class Message(AioRunMixin, metaclass=abc.ABCMeta):
    """
    A message class all simple mode messages are derived from.

    If the connection passed to the constructor is in blocking mode, then all
    methods on this class will be blocking too.

    Attributes:
        - `connection`: An instance of a connection.
        - `to_id`: Threema ID of the recipient.
    """
    async_functions = {
        'send',
    }

    def __init__(self, connection=None, to_id=None):
        super().__init__(blocking=connection.blocking)
        self._connection = connection.unwrap
        self.to_id = to_id

    @asyncio.coroutine
    @abc.abstractmethod
    def send(self):
        """
        Send a message.
        """


@aio_run_proxy_decorator
class TextMessage(Message):
    """
    Create and send a text message to a recipient.

    If the connection passed to the constructor is in blocking mode, then all
    methods on this class will be blocking too.

    Arguments / Attributes:
        - `connection`: An instance of a connection.
        - `id`: Threema ID of the recipient.
        - `phone`: Phone number of the recipient.
        - `email`: Email address of the recipient.
        - `text`: Message text.
    """
    async_functions = {
        'send',
    }

    def __init__(self, phone=None, email=None, text=None, **kwargs):
        super().__init__(**kwargs)
        self.phone = phone
        self.email = email
        self.text = text

    @asyncio.coroutine
    def send(self):
        """
        Send the created message.

        Return the ID of the sent message.
        """
        recipient = {
            'to': self.to_id,
            'phone': self.phone,
            'email': self.email
        }

        # Validate parameters
        if self._connection is None:
            raise MessageError('No connection set')
        if not any(recipient.values()):
            raise MessageError('No recipient specified')
        if sum([0 if to is None else 1 for to in recipient.values()]) > 1:
            raise MessageError('Only one recipient type can be used at the same time')
        if self.text is None:
            raise MessageError('Message text not specified')

        # Send message
        data = {key: value for key, value in recipient.items() if value is not None}
        data['text'] = self.text
        return (yield from self._connection.send_simple(**data))



"""
You can modify and use one of the functions below to test the gateway
service with your account.
"""
import logbook
import logbook.more

from threema.gateway import (
    Connection,
    GatewayError,
    util,
)
from threema.gateway.simple import TextMessage


def send_via_id(connection):
    """
    Send a message to a specific Threema ID.
    """
    message = TextMessage(
        connection=connection,
        to_id='ECHOECHO',
        text='Hello from the world of Python!'
    )
    return message.send()


def send_via_email(connection):
    """
    Send a message via an email address.
    """
    message = TextMessage(
        connection=connection,
        email='test@threema.ch',
        text='Hello from the world of Python!'
    )
    return message.send()


def send_via_phone(connection):
    """
    Send a message via a phone number.
    """
    message = TextMessage(
        connection=connection,
        phone='41791234567',
        text='Hello from the world of Python!'
    )
    return message.send()


def main():
    connection = Connection(
        identity='*YOUR_GATEWAY_THREEMA_ID',
        secret='YOUR_GATEWAY_THREEMA_ID_SECRET',
        verify_fingerprint=True,
        blocking=True,
    )
    try:
        with connection:
            send_via_id(connection)
            send_via_email(connection)
            send_via_phone(connection)
    except GatewayError as exc:
        print('Error:', exc)


if __name__ == '__main__':
    util.enable_logging(logbook.WARNING)
    log_handler = logbook.more.ColorizedStderrHandler()
    with log_handler.applicationbound():
        main()



import asyncio
import binascii
import enum

import aiohttp
import libnacl.encode
import libnacl.public

from .exception import (
    BlobServerError,
    CreditsServerError,
    GatewayKeyError,
    IDError,
    IDServerError,
    KeyServerError,
    MessageServerError,
    ReceptionCapabilitiesError,
    ReceptionCapabilitiesServerError,
)
from .key import Key
from .util import (
    AioRunMixin,
    aio_run_proxy_decorator,
    async_lru_cache,
    raise_server_error,
)

__all__ = (
    'ReceptionCapability',
    'Connection',
)


@enum.unique
class ReceptionCapability(enum.Enum):
    """
    The reception capability of a Threema ID.
    """
    text = 'text'
    image = 'image'
    video = 'video'
    audio = 'audio'
    file = 'file'


@aio_run_proxy_decorator
class Connection(AioRunMixin):
    """
    Container for the sender's Threema ID and the Threema Gateway
    secret. Can be applied to multiple messages for both simple and
    end-to-end mode.

    You should either use the `with` statement on this class or call
    :func:`~Connection.close` after you are done querying the Threema
    Gateway Service API. Be aware that the connection instance cannot be
    reused once it has been closed. This also applies to the `with`
    statement (e.g. the instance can be used in one `with` block only).
    A closed connection instance will raise :exc:`RuntimeError`
    indicating that the underlying HTTP session has been closed.

    The connection can work both in non-blocking (through asyncio) and
    blocking mode. If you want to use the API in a blocking way (which
    implicitly starts an event loop to process the requests), then
    instantiate this class with ``blocking=True``.

    Arguments:
        - `id`: Threema ID of the sender.
        - `secret`: Threema Gateway secret.
        - `key`: Private key of the sender. Only required for
          end-to-end mode.
        - `key_file`: A file where the private key is stored
          in. Can be used instead of passing the key directly.
        - `fingerprint`: A binary fingerprint of an DER-encoded TLS
          certificate. Will fall back to a stored fingerprint which will
          be invalid as soon as the certificate expires.
        - `verify_fingerprint`: Set to `True` if you want to verify the
          TLS certificate of the Threema Gateway Server by a
          fingerprint. (Recommended)
        - `blocking`: Whether to use a blocking API, without the need
          for an explicit event loop.
    """
    async_functions = {
        'get_public_key',
        'get_id',
        'get_reception_capabilities',
        'get_credits',
        'send_simple',
        'send_e2e',
        'upload',
        'download',
    }
    fingerprint = binascii.unhexlify(
        b'a6840a28041a1c43d90c21122ea324272f5c90c82dd64f52701f3a8f1a2b395b')
    urls = {
        'get_public_key': 'https://msgapi.threema.ch/pubkeys/{}',
        'get_id_by_phone': 'https://msgapi.threema.ch/lookup/phone/{}',
        'get_id_by_phone_hash': 'https://msgapi.threema.ch/lookup/phone_hash/{}',
        'get_id_by_email': 'https://msgapi.threema.ch/lookup/email/{}',
        'get_id_by_email_hash': 'https://msgapi.threema.ch/lookup/email_hash/{}',
        'get_reception_capabilities': 'https://msgapi.threema.ch/capabilities/{}',
        'get_credits': 'https://msgapi.threema.ch/credits',
        'send_simple': 'https://msgapi.threema.ch/send_simple',
        'send_e2e': 'https://msgapi.threema.ch/send_e2e',
        'upload_blob': 'https://msgapi.threema.ch/upload_blob',
        'download_blob': 'https://msgapi.threema.ch/blobs/{}'
    }

    def __init__(
            self, identity, secret,
            key=None, key_file=None,
            fingerprint=None, verify_fingerprint=False, blocking=False
    ):
        super().__init__(blocking=blocking)
        if fingerprint is None and verify_fingerprint:
            fingerprint = self.fingerprint
        connector = aiohttp.TCPConnector(fingerprint=fingerprint)
        self._session = aiohttp.ClientSession(connector=connector)
        self._key = None
        self._key_file = None
        self.id = identity
        self.secret = secret
        self.key = key
        self.key_file = key_file

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()

    def close(self):
        """
        Close the underlying :class:`aiohttp.ClientSession`.
        """
        self._session.close()

    @property
    def key(self):
        """Return the private key."""
        if self._key is None:
            raise GatewayKeyError("Sender's private key not specified")
        return self._key

    @key.setter
    def key(self, key):
        """Set the private key. The key will be decoded if required."""
        if isinstance(key, str):
            key = Key.decode(key, Key.Type.private)
        self._key = key

    @property
    def key_file(self):
        """Get the path of the private key file."""
        return self._key_file

    @key_file.setter
    def key_file(self, key_file):
        """Set the private key by reading it from a file."""
        if key_file is not None:
            with open(key_file) as file:
                self.key = file.readline().strip()
        self._key_file = key_file

    @asyncio.coroutine
    @async_lru_cache(maxsize=1024, expiration=60 * 60)
    def get_public_key(self, id_):
        """
        Get the public key of a Threema ID.

        Arguments:
            - `id_`: A Threema ID.

        Return a :class:`libnacl.public.PublicKey` for a Threema ID.
        """
        response = yield from self._get(self.urls['get_public_key'].format(id_))
        if response.status == 200:
            text = yield from response.text()
            key = libnacl.encode.hex_decode(text)
            return libnacl.public.PublicKey(key)
        else:
            yield from raise_server_error(response, KeyServerError)

    @asyncio.coroutine
    @async_lru_cache(maxsize=1024, expiration=60 * 60)
    def get_id(self, **mode):
        """
        Get a user's Threema ID.

        Use **only one** of the arguments described below.

        Arguments:
            - `phone`: A phone number in E.164 format without the
              leading `+`.
            - `phone_hash`: An HMAC-SHA256 hash of an E.164 phone
              number without the leading `+`.
            - `email`: A lowercase email address.
            - `email_hash`: An HMAC-SHA256 hash of a lowercase and
              whitespace-trimmed email address.

        Return the Threema ID.
        """
        modes = {
            'phone': 'get_id_by_phone',
            'phone_hash': 'get_id_by_phone_hash',
            'email': 'get_id_by_email',
            'email_hash': 'get_id_by_email_hash'
        }

        # Check mode
        if len(set(mode) - set(modes)) > 0:
            raise IDError('Unknown mode selected: {}'.format(set(mode)))
        mode_length = len(mode)
        if mode_length > 1 or mode_length == 0:
            raise IDError('Use (only) one of the possible modes to get a Threema ID')

        # Select mode and start request
        mode, value = mode.popitem()
        response = yield from self._get(self.urls[modes[mode]].format(value))
        if response.status == 200:
            return (yield from response.text())
        else:
            yield from raise_server_error(response, IDServerError)

    @asyncio.coroutine
    @async_lru_cache(maxsize=128, expiration=5 * 60)
    def get_reception_capabilities(self, id_):
        """
        Get the reception capabilities of a Threema ID.

        Arguments:
            - `id_`: A Threema ID.

        Return a set containing items from :class:`ReceptionCapability`.
        """
        get_coroutine = self._get(self.urls['get_reception_capabilities'].format(id_))
        response = yield from get_coroutine
        if response.status == 200:
            try:
                text = yield from response.text()
                return {ReceptionCapability(capability.strip())
                        for capability in text.split(',')}
            except ValueError as exc:
                yield from response.release()
                raise ReceptionCapabilitiesError('Invalid reception capability') from exc
        else:
            yield from raise_server_error(response, ReceptionCapabilitiesServerError)

    @asyncio.coroutine
    def get_credits(self):
        """
        Return the number of credits left on the account.
        """
        response = yield from self._get(self.urls['get_credits'])
        if response.status == 200:
            text = yield from response.text()
            return int(text)
        else:
            yield from raise_server_error(response, CreditsServerError)

    @asyncio.coroutine
    def send_simple(self, **data):
        """
        Send a message by using the simple mode.

        Arguments:
            - `data`: A dictionary containing POST data.

        Return the ID of the message.
        """
        return (yield from self._send(self.urls['send_simple'], data))

    @asyncio.coroutine
    def send_e2e(self, **data):
        """
        Send a message by using the end-to-end mode.

        Arguments:
            - `data`: A dictionary containing POST data.

        Return the ID of the message.
        """
        return (yield from self._send(self.urls['send_e2e'], data))

    @asyncio.coroutine
    def upload(self, data):
        """
        Upload a blob.

        Arguments:
            - `data`: Binary data.

        Return the hex-encoded ID of the blob.
        """
        return (yield from self._upload(self.urls['upload_blob'], data))

    @asyncio.coroutine
    def download(self, blob_id):
        """
        Download a blob.

        Arguments:
            - `id`: The hex-encoded blob ID.

        Return a :class:`asyncio.StreamReader` instance.
        """
        response = yield from self._get(self.urls['download_blob'].format(blob_id))
        if response.status == 200:
            return response.content
        else:
            yield from raise_server_error(response, BlobServerError)

    @asyncio.coroutine
    def _get(self, *args, **kwargs):
        """
        Wrapper for :func:`requests.get` that injects the connection's
        Threema ID and its secret.

        Return a :class:`aiohttp.ClientResponse` instance.
        """
        kwargs.setdefault('params', {})
        kwargs['params'].setdefault('from', self.id)
        kwargs['params'].setdefault('secret', self.secret)
        return (yield from self._session.get(*args, **kwargs))

    @asyncio.coroutine
    def _send(self, url, data):
        """
        Send a message.

        Arguments:
            - `url`: URL for the request.
            - `data`: A dictionary containing POST data.

        Return the ID of the message.
        """
        # Inject Threema ID and secret
        data.setdefault('from', self.id)
        data.setdefault('secret', self.secret)

        # Send message
        response = yield from self._session.post(url, data=data)
        if response.status == 200:
            return (yield from response.text())
        else:
            yield from raise_server_error(response, MessageServerError)

    @asyncio.coroutine
    def _upload(self, url, data):
        """
        Upload a blob.

        Arguments:
            - `data`: Binary data.

        Return the ID of the blob.
        """
        # Inject Threema ID and secret
        params = {'from': self.id, 'secret': self.secret}

        # Prepare multipart encoded file
        files = {'blob': data}

        # Send message
        response = yield from self._session.post(url, params=params, data=files)
        if response.status == 200:
            return (yield from response.text())
        else:
            yield from raise_server_error(response, BlobServerError)



"""
You can modify and use one of the lines below to test the lookup
functionality of the gateway service.
"""
import asyncio

import logbook
import logbook.more

from threema.gateway import (
    Connection,
    GatewayError,
    util,
)
from threema.gateway.key import Key


@asyncio.coroutine
def main():
    connection = Connection(
        identity='*YOUR_GATEWAY_THREEMA_ID',
        secret='YOUR_GATEWAY_THREEMA_ID_SECRET',
        verify_fingerprint=True,
    )
    try:
        with connection:
            print((yield from connection.get_credits()))
            print((yield from connection.get_id(phone='41791234567')))
            hash_ = 'ad398f4d7ebe63c6550a486cc6e07f9baa09bd9d8b3d8cb9d9be106d35a7fdbc'
            print((yield from connection.get_id(phone_hash=hash_)))
            print((yield from connection.get_id(email='test@threema.ch')))
            hash_ = '1ea093239cc5f0e1b6ec81b866265b921f26dc4033025410063309f4d1a8ee2c'
            print((yield from connection.get_id(email_hash=hash_)))
            key = (yield from connection.get_public_key('ECHOECHO'))
            print(Key.encode(key))
            print((yield from connection.get_reception_capabilities('ECHOECHO')))
    except GatewayError as exc:
        print('Error:', exc)


if __name__ == '__main__':
    util.enable_logging(logbook.WARNING)
    log_handler = logbook.more.ColorizedStderrHandler()
    with log_handler.applicationbound():
        loop = asyncio.get_event_loop()
        loop.run_until_complete(main())
        loop.close()
